'''
Raven RSS Reader
Powered by ThinkSQL RDBMS

(c)Copyright 2004 ThinkSQL Ltd. All rights reserved.


Uses "Universal feed parser" and "Ultra-liberal feed finder" from Mark Pilgrim
Visit http://diveintomark.org/projects/feed_parser/ for the latest version

Uses "db_row" from Kevin Jacobs (jacobs@theopalgroup.com)


Todo:
    Add html progress during update & improve error reporting
    Improve admin form: allow purge maintenance, stylesheet selection, thread-pool sizing etc.
    Allow article flagging (moving/copying into 'saved' group)
    Auto-purge
    Show unread total per group and allow group contract/expand
    Tag feeds as contributing to a summary page
    Show favicon/image against each feed
'''

__version__ = "0.103 beta"
__copyright__ = "(c)Copyright 2004 ThinkSQL Ltd. All rights reserved."
__author__ = "ThinkSQL Ltd"

import BaseHTTPServer
import cgi
import sys
import re
import socket
import SocketServer

from datetime import datetime, timedelta
import time
from string import replace
from random import randrange

import ThinkSQL
from db_row import IMetaRow

import feedparser
import feedfinder

import threading

import os
import locale
import csv
import errno

locale.setlocale(locale.LC_ALL, '')

ISO_DATE='%Y-%m-%d %H:%M:%S'
USER_DATE='%c'
MAX_THREADS=10  #thread-pool size for parallel feed parsing

def strEncode(s):
    try:
        return s.encode('latin-1', 'replace')
    except:
        try:
            return unicode(s, 'latin-1').encode('latin-1', 'replace')
        except:
            return '?'

def strSafe(s):
    return replace(strEncode(s),"'","''")

def debug(s):
    global debugSwitch
    if debugSwitch:
        print s
        
class parser(threading.Thread):
    '''Parses a feed in a separate thread'''
    def __init__(self, id, name, url, etag, lastmod):
        threading.Thread.__init__(self)
        self.id = id
        self.name = name
        self.url = url
        self.etag = etag
        self.lastmod = lastmod
        debug('Creating thread for %s' % name)
        
    def run(self):
        try:
            self.feed=feedparser.parse(self.url, self.etag, self.lastmod)
            debug('  Parsed %s' % self.name)
            #todo If we have a multi-connection ThinkSQL licence, update the database here with a parallel connection
        except Exception, e:
            print '  Failed reading %s (%s)' % (self.name, str(e))
            self.feed=None            
        #when this method returns, the thread exits


class feedReader:
    def __init__(self, conn):
        self.conn=conn
        self.parsers=None
        self.lock=threading.RLock()
        self.lock.acquire()
        try:
            results1=self.conn.cursor()
            results1.execute('SELECT * FROM SITES')
            SITES = IMetaRow(results1.description)
            self.siteId=None
            self.siteName=''
            for s in results1:
                self.siteId=SITES(s).fields.SITE_ID
                self.siteName=SITES(s).fields.SITE_NAME
                self.port=SITES(s).fields.PORT
                self.style=SITES(s).fields.STYLE
                self.summary_size=SITES(s).fields.SUMMARY_SIZE
                self.article_live_days=SITES(s).fields.ARTICLE_LIVE_DAYS
                self.article_purge_days=SITES(s).fields.ARTICLE_PURGE_DAYS
                self.article_limit=30   #todo put in SITES
            results1.close()
        finally:
            self.lock.release()

    def refresh(self):
        debug('Refreshing...')
        self.parsers=[]
        
        self.lock.acquire()
        try:
            #Retrieve results
            results1=self.conn.cursor()
            try:
                results1.execute('SELECT * FROM GROUPS ORDER BY GROUP_NAME')
                GROUPS = IMetaRow(results1.description)

                for g in results1:
                    debug(GROUPS(g).fields.GROUP_NAME)

                    results2=self.conn.cursor()
                    results2.execute("SELECT * FROM FEEDS WHERE GROUP_ID=%d ORDER BY FEED_NAME" % GROUPS(g).fields.GROUP_ID)
                    FEEDS = IMetaRow(results2.description)

                    #Launch multiple feed parser threads in parallel
                    for f in results2:
                        debug('  '+FEEDS(f).fields.FEED_NAME)
                        lastmod=None
                        if FEEDS(f).fields.FEED_MODIFIED:
                            lastmod=FEEDS(f).fields.FEED_MODIFIED.timetuple()
                            
                        while threading.activeCount() >= MAX_THREADS:
                            time.sleep(1)   #wait for a thread to become available
                        p = parser(FEEDS(f).fields.FEED_ID, FEEDS(f).fields.FEED_NAME, FEEDS(f).fields.FEED_URL, FEEDS(f).fields.FEED_ETAG, lastmod)
                        self.parsers.append(p)
                        p.start()
                    results2.close()
            finally:
                results1.close()
                    
            #Process results (serial updates to the server while we have a single-connection licence)
            for p in self.parsers:                                               
                debug('Waiting for thread %s' % p.name)
                p.join() #wait for thread to finish
            
                feed=p.feed
                
                if feed: 
                    debug('Processing %s' % p.name)
                    if feed.has_key('modified'):
                        base=datetime(feed['modified'][0],feed['modified'][1],feed['modified'][2],feed['modified'][3],feed['modified'][4],feed['modified'][5])
                    else:
                        base=datetime.today()
                    seq=timedelta(seconds=0)
                    for article in feed['items']:
                        date=base - seq
                        if article.has_key('date_parsed'):
                            date=datetime(article['date_parsed'][0],article['date_parsed'][1],article['date_parsed'][2],article['date_parsed'][3],article['date_parsed'][4],article['date_parsed'][5])
                        else:
                            seq=seq+timedelta(seconds=1)    #retain original order: first = most recent
     
                        if article.has_key('link'):
                            results3=self.conn.cursor()
                            results3.execute("SELECT * FROM ARTICLES WHERE FEED_ID=%d AND ARTICLE_LINK='%s'" % (p.id, strSafe(article['link'])))
                            ARTICLES = IMetaRow(results3.description)
                            a=results3.fetchone()
                            if a:
                                #already exists
                                pass
                            else:
                                if article.has_key('description') and article['description']:
                                    try:
                                        description=strEncode(article['description'])
                                    except Exception, e:
                                        description='?: %s' % str(e)
                                else:
                                    description=' '
                                exec1=self.conn.cursor()
                                exec1.execute("INSERT INTO ARTICLES (FEED_ID, ARTICLE_LINK, ARTICLE_TITLE, ARTICLE_DESCRIPTION, ARTICLE_DATE, ARTICLE_STATUS) VALUES (?, ?, ?, ?, ?, ?)" , (p.id, strEncode(article['link']), strEncode(article['title']), description, date.strftime(ISO_DATE), STATUS_UNREAD))
                                exec1.close()
                                debug('    Inserted %s' % article['link'])
                            results3.close()
                        #else invalid article
     
                    if feed.has_key('modified'):
                        debug('    Modified '+str(feed['modified']))
                        if feed.has_key('etag') and feed['etag']:
                            etag=strEncode(feed['etag'])
                        else:
                            etag=None
                        exec1=self.conn.cursor()
                        try:
                            exec1.execute("UPDATE FEEDS SET FEED_MODIFIED=?, FEED_ETAG=? WHERE FEED_ID=?", (datetime(feed['modified'][0],feed['modified'][1],feed['modified'][2],feed['modified'][3],feed['modified'][4],feed['modified'][5]), etag, p.id))
                        except:
                            #todo print exception
                            pass
                        exec1.close()
                    else:
                        debug('    '+'Not modified since '+str(p.lastmod))
     
                    self.conn.commit()
                #else failed to parse
            debug('Refreshed')
        finally:
            self.lock.release()

#ThinkSQL Server Parameters
DEFAULT_HOST='localhost'
DEFAULT_PORT=9075
DEFAULT_USER='RAVEN'

STATE_DISCONNECTED=00
STATE_CONNECTED=10

STATUS_UNREAD='U'
STATUS_READ='R'

class ServerSession:
    def __init__(self):
        self.id=randrange(1,2000000000)  #todo retry while exists in sessions already
        self.conn=None
        self.state=STATE_DISCONNECTED
        self.host=DEFAULT_HOST
        self.port=DEFAULT_PORT
        self.username=DEFAULT_USER
        self.password=''
        self.page=None
        self.feedId=None
        self.articleId=None
        self.feedlist=None

        self.reader=None

        self.lock=threading.RLock()

    def __del__(self):
        if self.state==STATE_CONNECTED:
            self.conn.close()

    def __str__(self):
        return str(self.id)

sessions={}

#########################

PAGE_HEAD='''<html>
<head>
<title>Raven</title>
<link rel="stylesheet" type="text/css" href="default.css" />

<script type="text/javascript">
<!--
function toggle(id){
    ul = "ul_" + id;
    uldesc = "ul_desc" + id;
    img = "img_" + id;
    ulElement = document.getElementById(ul);
    ulDescElement = document.getElementById(uldesc);
    imgElement = document.getElementById(img);
    if (ulElement){
        if (ulElement.className == 'closed'){
            ulElement.className = "open";
            ulDescElement.className = "closed";
            imgElement.src = "open.jpg";
            }else{
            ulElement.className = "closed";
            ulDescElement.className = "open";
            imgElement.src = "closed.jpg";
            }
        }
    }
//-->
</script>
<META HTTP-EQUIV="Pragma" CONTENT="no-cache">
<META HTTP-EQUIV="Expires" CONTENT="-1">
</head>

<body>
'''

FORM_HEAD='''
<form method="GET" action="http://%s">
<input type="hidden" name="session_id" value="%d">
<input type="hidden" name="page" value="%s">
'''

FORM_FOOT='''
</form>
'''

PAGE_FOOT='''
</body>
</html>
'''

FORM_ERROR='''
<H1>Error</H1>
  %s
'''

FORM_GROUP_BOTTOM='''<br /><br /><div align="center">Powered by <a href="http://www.thinksql.co.uk">ThinkSQL</a></div>'''

FORM_HEADER_TOP='''<div id="top">
<H1><img src="raven.gif"/> %(title)s</H1><div class="date">%(date)s</div>
</div>'''

FORM_MAIN_TOP=FORM_HEADER_TOP+'''
<div id="left">
<h3>Feeds</h3>
<input type=submit name="Command" value="Update All"><br />
%(feeds)s
<input type=submit name="Command" value="Add Feed..."><br /><br />
<input type=submit name="Command" value="Admin...">
'''+FORM_GROUP_BOTTOM+'''</div>

<div id="middle">
'''

FORM_MAIN_ARTICLE='''<h4><a href="#%d" onclick="toggle('item%d')"><img id="img_item%d" src="closed.jpg" /></a>
 <a name="%d" href="%s">%s</a></h4><div class="date">%s</div>
<div id="ul_descitem%d" class="open">%s</div><div id="ul_item%d" class="closed">%s</div>'''

FORM_MAIN_ARTICLE_OPEN='''<h4>
 <a href="%s">%s</a></h4><div class="date">%s</div>
%s'''

FORM_MAIN_BOTTOM='''
</div>
'''

FORM_GROUP='''<h4><img src="open.jpg" /> %s</h4>'''
FORM_ACTIVE_FEED='''&nbsp;&nbsp;&nbsp;<img src="bullet.gif" /><a href="%s"><i>%s%s</i></a><br />'''
FORM_FEED='''&nbsp;&nbsp;&nbsp;<img src="bullet.gif" /><a href="%s">%s%s</a><br />'''

FORM_ADDFEED_TOP=FORM_HEADER_TOP+'''
<div id="left">
<h3>Feeds</h3>
%(feeds)s
<p><input type=submit name="Command" value="Main"></p>
'''+FORM_GROUP_BOTTOM+'''</div>

<div id="middle">
'''

FORM_ADDFEED_MAIN='''<h2>%s</h2><br />
<table border="0" width=100%% style="PADDING-LEFT: 5px; PADDING-BOTTOM: 5px; PADDING-TOP: 5px; BACKGROUND-COLOR: #EEE8AA">
<tr><td>Name:</td><td><INPUT TYPE=TEXT NAME="feedname" value="%s" SIZE=50></td></tr>
<tr><td>URL:</td><td><INPUT TYPE=TEXT NAME="feedurl" SIZE=50 VALUE="%s"> <input type=submit name="Command" value="Find"> <img src="rss.gif" /></td></tr>
</table>
<p align=center><input type=submit name="Command" value="Add"></p>
'''

FORM_ADMIN_MAIN='''<h2>%s</h2><br />
<table border="0" width=100%% style="PADDING-LEFT: 5px; PADDING-BOTTOM: 5px; PADDING-TOP: 5px; BACKGROUND-COLOR: #B0C4DE">
<tr><th width=150>Current Feed</th><th></th></tr>
<tr><td></td><td><input type=submit name="Command" value="Update Feed"> <input type=submit name="Command" value="Move Feed..."> <input type=submit name="Command" value="Delete Feed"></td></tr>
<tr><td>Name:</td><td><INPUT TYPE=TEXT NAME="feedname" SIZE=60 VALUE="%s"></td></tr>
<tr><td>URL:</td><td><INPUT TYPE=TEXT NAME="feedurl" SIZE=60 VALUE="%s"></td></tr>
<tr><td>Collapse after:</td><td><INPUT TYPE=TEXT NAME="feedsummarysize" SIZE=5 MAXLENGTH=5 VALUE="%s"> characters (blank = server default)</td></tr>
</table>
<br />
<table border="0" width=100%% style="PADDING-LEFT: 5px; PADDING-BOTTOM: 5px; PADDING-TOP: 5px; BACKGROUND-COLOR: #4682B4">
<tr><th width=150>Current Group</th><th></th></tr>
<tr><td></td><td><input type=submit name="Command" value="Update Group"> <input type=submit name="Command" value="Delete Group"> <input type=submit name="Command" value="Export Group"></td></tr>
<tr><td>Name:</td><td><INPUT TYPE=TEXT NAME="groupname" SIZE=60 VALUE="%s"></td></tr>
</table>
<br />
<table border="0" width=100%% style="PADDING-LEFT: 5px; PADDING-BOTTOM: 5px; PADDING-TOP: 5px; BACKGROUND-COLOR: #EEE8AA">
<tr><th width=150>New Group</th><th></th></tr>
<tr><td></td><td><input type=submit name="Command" value="Add Group"> <input type=submit name="Command" value="Import Group"></td></tr>
<tr><td>Name:</td><td><INPUT TYPE=TEXT NAME="newgroupname" SIZE=60></td></tr>
</table>
<br />
<table border="0" width=100%% style="PADDING-LEFT: 5px; PADDING-BOTTOM: 5px; PADDING-TOP: 5px; BACKGROUND-COLOR: #FF9966">
<tr><th width=150>Server Configuration</th><th></th></tr>
<tr><td></td><td><input type=submit name="Command" value="Update Server"></td></tr>
<tr><td>Purge older than:</td><td><INPUT TYPE=TEXT NAME="serverpurgedays" SIZE=4 MAXLENGTH=4 VALUE="%s"> days</tr>
</table>
<br />
<table border="0" width=100%% style="PADDING-LEFT: 5px; PADDING-BOTTOM: 5px; PADDING-TOP: 5px; BACKGROUND-COLOR: #FFD700">
<tr><th width=150>Server</th><th></th></tr>
<tr><td></td><td><input type=submit name="Command" value="Statistics"></td></tr>
<tr><td></td><td><input type=submit name="Command" value="Purge"></td></tr>
<tr><td></td><td><input type=submit name="Command" value="Shutdown"></td></tr>
</table>
'''

FORM_MOVEGROUP='''<h4><a href="%s">%s</a></h4>'''

FORM_MOVEFEED_MAIN='''<h2>%(title)s</h2><br />
<table border="0" width=100%% style="PADDING-LEFT: 5px; PADDING-BOTTOM: 5px; PADDING-TOP: 5px; BACKGROUND-COLOR: #B0C4DE">
<tr><td>
%(groups)s
</td></tr>
</table>
<p align=center><input type=submit name="Command" value="Cancel"></p>
'''

FORM_STATS='''<tr><td>%s</td><td align=right>%d</td><td align=right>%s</td><td align=right>%d</td></tr>'''

FORM_STATS_MAIN='''<h2>%(title)s</h2><br />
<table border="0" width=100%% style="PADDING-LEFT: 5px; PADDING-BOTTOM: 5px; PADDING-TOP: 5px; BACKGROUND-COLOR: #B0C4DE">
<tr><th>Feed</th><th>Articles</th><th>Earliest Article</th><th>Content size (bytes)</th></tr>
%(stats)s
</table>
'''
#<p align=center><input type=submit name="Command" value="Main"></p>

#########################

BaseRHClass = BaseHTTPServer.BaseHTTPRequestHandler
class testRequestHandler(BaseRHClass):

    def showFormHead(self,session,page):
        self.wfile.write( PAGE_HEAD )
        global httpd
        formreturn='%s:%s' % (str(self.client_address[0]), httpd.server_port) 
        self.wfile.write( FORM_HEAD % (formreturn, session.id, page) )

    def showFormFoot(self,session):
        self.wfile.write( FORM_FOOT )
        self.wfile.write( PAGE_FOOT )

    def showError(self,session,message):
        self.showFormHead(session,'showError')
        self.wfile.write( FORM_ERROR % message)
        self.showFormFoot(session)

    def getGroupMenu(self, session, feedId=None):
        '''Returns tuple of (feedId, feedName, feedSummary_size, groupMenuTextList)'''
        feedName=''
        feedSummary_size=session.reader.summary_size    #default
        feeds=[]

        if not session.feedlist:
            #Refresh cache: list of [group_name, feed_id, feed_name, feed_summary_size, total_unread]
            session.lock.acquire()
            try:
                session.feedlist=[]
                this_group=None
                results1=session.conn.cursor()
                results1.execute('SELECT * FROM GROUPS ORDER BY GROUP_NAME')
                GROUPS = IMetaRow(results1.description)
                for g in results1:
                    if GROUPS(g).fields.GROUP_NAME<>this_group:
                        #New group
                        this_group=GROUPS(g).fields.GROUP_NAME

                    base=datetime.today()
                    seq=timedelta(days=session.reader.article_live_days)
                    base=base-seq
                    cutoff_date=base.strftime(ISO_DATE)
                    results2=session.conn.cursor()
                    results2.execute("""SELECT *, (SELECT COUNT(*) FROM ARTICLES
                                                   WHERE FEED_ID=FEEDS.FEED_ID
                                                   AND ARTICLE_STATUS='%s'
                                                   AND ARTICLE_DATE>TIMESTAMP'%s') AS TOTAL_UNREAD
                                        FROM FEEDS
                                        WHERE GROUP_ID=%d
                                        ORDER BY FEED_NAME""" % (STATUS_UNREAD, cutoff_date, GROUPS(g).fields.GROUP_ID))
                    FEEDS = IMetaRow(results2.description)
                    for f in results2:
                        session.feedlist.append( [this_group, FEEDS(f).fields.FEED_ID, FEEDS(f).fields.FEED_NAME, FEEDS(f).fields.FEED_SUMMARY_SIZE, FEEDS(f).fields.TOTAL_UNREAD] )

                    results2.close()

                results1.close()
            finally:
                session.lock.release()

        #Use cached feedlist to build string 
        this_group=None
        for f in session.feedlist:
            if f[0]<>this_group:
                #New group
                this_group=f[0]
                feeds.append(FORM_GROUP % (this_group))

            if not feedId:
                feedId=f[1]

            if f[4]==0:
                showUnread=''
            else:
                showUnread=' (%d)' % f[4]

            if int(feedId)==f[1]:
                feedName=f[2]
                if f[3]:
                    feedSummary_size=f[3]
                feeds.append(FORM_ACTIVE_FEED % ('/?session_id=%d&page=%s&Command=Feed&FeedId=%d' % (session.id, 'actionMainForm', f[1]), f[2], showUnread))
            else:
                feeds.append(FORM_FEED % ('/?session_id=%d&page=%s&Command=Feed&FeedId=%d' % (session.id, 'actionMainForm', f[1]), f[2], showUnread))

        feeds.append('<br />')

        return (feedId, feedName, feedSummary_size, feeds)

    def setGroupMenuUnread(self, session, feedId=None):
        '''Resets unread count when moving off a feed link'''
        for f in session.feedlist:
            if f[1]==int(feedId) and f[4]<>0:
                session.lock.acquire()
                try:
                    results1=session.conn.cursor()
                    try:
                        results1.execute("UPDATE ARTICLES SET ARTICLE_STATUS='%s' WHERE FEED_ID=%d" % (STATUS_READ, int(feedId)))
                    except:
                        #todo print exception
                        pass
                    results1.close()
                    
                    session.conn.commit()
                finally:
                    session.lock.release()
                    
                f[4]=0
                

    def getCommand(self,session,fields):
        result=None
        if fields.has_key('Command'):
            result=fields['Command']
            if len(result)>1 and result[-1]=='/':
                result=result[:-1]
        return result

    def showMainForm(self, session, feedId=None, articleId=None):
        self.showFormHead(session,'actionMainForm')
        
        debug('Starting %s' % threading.currentThread())

        gm=self.getGroupMenu(session, feedId)
        feedId=gm[0]
        feedName=gm[1]
        feedSummary_size=gm[2]
        feeds=gm[3]

        feeds.append('<br />')

        session.feedId=feedId
        session.articleId=articleId

        self.wfile.write( FORM_MAIN_TOP % {'title':session.reader.siteName, 'date':datetime.today().strftime(USER_DATE), 'feeds':''.join(feeds)} )

        self.wfile.write('<h2>%s</h2>' % feedName)

        base=datetime.today()
        seq=timedelta(days=session.reader.article_live_days)
        base=base-seq
        cutoff_date=base.strftime(ISO_DATE)
        session.lock.acquire()
        try:
            results1=session.conn.cursor()
            results1.execute("""SELECT * FROM ARTICLES
                                WHERE FEED_ID=%d
                                AND ARTICLE_DATE>TIMESTAMP'%s'
                                ORDER BY ARTICLE_DATE DESC, ARTICLE_TITLE""" % (int(feedId), cutoff_date))
            ARTICLES = IMetaRow(results1.description)

            rn=0
            lastDate=None
            for a in results1:
                rn += 1
                if not articleId:
                    articleId=ARTICLES(a).fields.ARTICLE_ID

                if lastDate and ARTICLES(a).fields.ARTICLE_DATE and ARTICLES(a).fields.ARTICLE_DATE.day<>lastDate.day:
                    self.wfile.write('<hr>')
                lastDate=ARTICLES(a).fields.ARTICLE_DATE

                desc = re.sub('<!--.*?-->', '', ARTICLES(a).fields.ARTICLE_DESCRIPTION) #Remove comments first, or '>' in
                                                      #comments will be interpreted as
                                                      #end of (comment) tag.
                desc = re.sub('<.*?>', '', desc)

                if len(desc)>=feedSummary_size:
                    desc=desc[:feedSummary_size]+'...'
                    self.wfile.write( FORM_MAIN_ARTICLE % (ARTICLES(a).fields.ARTICLE_ID, ARTICLES(a).fields.ARTICLE_ID, ARTICLES(a).fields.ARTICLE_ID,
                                                           ARTICLES(a).fields.ARTICLE_ID, ARTICLES(a).fields.ARTICLE_LINK, ARTICLES(a).fields.ARTICLE_TITLE, ARTICLES(a).fields.ARTICLE_DATE.strftime(USER_DATE),
                                                           ARTICLES(a).fields.ARTICLE_ID, desc, ARTICLES(a).fields.ARTICLE_ID,
                                                           ARTICLES(a).fields.ARTICLE_DESCRIPTION) )
                else:
                    self.wfile.write( FORM_MAIN_ARTICLE_OPEN % (ARTICLES(a).fields.ARTICLE_LINK, ARTICLES(a).fields.ARTICLE_TITLE, ARTICLES(a).fields.ARTICLE_DATE.strftime(USER_DATE),
                                                                ARTICLES(a).fields.ARTICLE_DESCRIPTION) )


                if rn>session.reader.article_limit:
                    break

            self.wfile.write( FORM_MAIN_BOTTOM )

            results1.close()

            self.setGroupMenuUnread(session, feedId)
            
            debug('Ending %s' % threading.currentThread())
        finally:
            debug('Finally %s' % threading.currentThread())
            session.lock.release()

        self.showFormFoot(session)

    def actionMainForm(self,session,fields):
        command=self.getCommand(session,fields)
        feedId=None
        articleId=None

        if command == "Update All":
            session.reader.refresh()
            session.feedlist=None
            self.showMainForm(session, session.feedId, session.articleId)
        elif command == "Add Feed...":
            self.showAddFeedForm(session, session.feedId)
        elif command == "Admin...":
            self.showAdminForm(session, session.feedId)
        elif command == "Feed":
            if fields.has_key('FeedId'):
                fId=fields['FeedId']
            self.showMainForm(session, fId, None)
        elif command == "Article":
            if fields.has_key('ArticleId'):
                aId=fields['ArticleId']
            self.showMainForm(session, session.feedId, aId)
        else:
            self.showError(session,"Unknown command: "+command)

    def showAddFeedForm(self, session, feedId=None, defaultURL='http://', defaultName=''):
        self.showFormHead(session,'actionAddFeedForm')

        gm=self.getGroupMenu(session, feedId)
        feedId=gm[0]
        feedName=gm[1]
        feedSummary_size=gm[2]
        feeds=gm[3]

        feeds.append('<br />')

        self.wfile.write( FORM_ADDFEED_TOP % {'title':session.reader.siteName, 'date':datetime.today().strftime(USER_DATE), 'feeds':''.join(feeds)} )

        self.wfile.write( FORM_ADDFEED_MAIN % ('Add New RSS/Atom Feed', defaultName, defaultURL) )

        self.showFormFoot(session)

    def actionAddFeedForm(self,session,fields):
        command=self.getCommand(session,fields)

        if command == "Main":
            self.showMainForm(session, session.feedId, session.articleId)
        elif command == "Find":
            fURL=None
            if fields.has_key('feedurl'):
                fURL=fields['feedurl']
                if fURL.startswith('http://'*2):
                    fURL = fURL[7:]
                found=feedfinder.getFeeds(fURL)
            else:
                found = False
            if found:
                fName=feedparser.parse(found[0])['channel']['title']              
                fURL=found[0]
            else:
                fName=''

            self.showAddFeedForm(session, session.feedId, fURL, fName)
        elif command == "Add":
            debug("Adding feed")
            fName = fields.get('feedname','')
            fURL = fields.get('feedurl','')
            fsummarysize = fields.get('feedsummarysize','')
            assert fName.strip(), "Feeds must have a name"
            assert fURL.strip(), "Feeds must have a URL"
            self.AddFeed(session, fName, fURL)
            self.showMainForm(session, session.feedId, session.articleId)
        #todo else?

    def AddFeed(self,session,feedName,feedURL):
        assert session.feedId, 'Needs current feed/group'   #todo else create initial group...

        session.lock.acquire()
        try:
            results1=session.conn.cursor()
            results1.execute('SELECT * FROM FEEDS WHERE FEED_ID=%d' % int(session.feedId))
            FEEDS = IMetaRow(results1.description)
            f=results1.fetchone()
            if f:
                exec1=session.conn.cursor()
                exec1.execute("INSERT INTO FEEDS (GROUP_ID, FEED_NAME, FEED_URL) VALUES (?, ?, ?)" , (FEEDS(f).fields.GROUP_ID, strEncode(feedName), strEncode(feedURL)))
                exec1.close()

                debug('Added %s (%s)' % (feedName, feedURL))

                session.conn.commit()
            #todo else error!
            results1.close()
        finally:
            session.lock.release()

        session.feedlist=None


    def UpdateFeed(self,session,feedName,feedURL,feedSummarysize):
        assert session.feedId, 'Needs current feed/group'   #todo else create initial group...

        session.lock.acquire()
        try:
            if feedSummarysize:
                try:
                    feedSummarysize=int(feedSummarysize)
                except:
                    feedSummarysize=None
            exec1=session.conn.cursor()
            if feedSummarysize:
                exec1.execute("UPDATE FEEDS SET FEED_NAME=?, FEED_URL=?, FEED_SUMMARY_SIZE=? WHERE FEED_ID=?" , (strEncode(feedName), strEncode(feedURL), feedSummarysize, session.feedId))
            else:                
                exec1.execute("UPDATE FEEDS SET FEED_NAME=?, FEED_URL=?, FEED_SUMMARY_SIZE=NULL WHERE FEED_ID=?" , (strEncode(feedName), strEncode(feedURL), session.feedId))
            exec1.close()

            debug('Updated feed %d to %s (%s)' % (int(session.feedId), feedName, feedURL))

            session.conn.commit()
        finally:
            session.lock.release()

        session.feedlist=None


    def showMoveFeedForm(self,session):
        self.showFormHead(session,'actionMoveFeedForm')

        gm=self.getGroupMenu(session, session.feedId)
        feedId=gm[0]
        feedName=gm[1]
        feedSummary_size=gm[2]
        feeds=gm[3]

        feeds.append('<br />')

        self.wfile.write( FORM_ADDFEED_TOP % {'title':session.reader.siteName, 'date':datetime.today().strftime(USER_DATE), 'feeds':''.join(feeds)} )

        groups=[]
        session.lock.acquire()
        try:
            results1=session.conn.cursor()
            results1.execute('SELECT * FROM GROUPS ORDER BY GROUP_NAME')
            GROUPS = IMetaRow(results1.description)
            for g in results1:
                groups.append(FORM_MOVEGROUP % ('/?session_id=%d&page=%s&Command=MoveFeed&GroupId=%d' % (session.id, 'actionMoveFeedForm', GROUPS(g).fields.GROUP_ID), GROUPS(g).fields.GROUP_NAME))

            results1.close()
        finally:
            session.lock.release()

        self.wfile.write( FORM_MOVEFEED_MAIN % {'title':'Move %s to Another Group' % feedName, 'groups':''.join(groups)} )

        self.showFormFoot(session)


    def actionMoveFeedForm(self,session,fields):
        command=self.getCommand(session,fields)

        if command == "Main":
            self.showMainForm(session, session.feedId, session.articleId)
        elif command == "Cancel":
            self.showMainForm(session, session.feedId, session.articleId)
        elif command == "MoveFeed":
            if fields.has_key('GroupId'):
                gId=fields['GroupId']
            self.MoveFeed(session,gId)
            self.showMainForm(session, session.feedId, session.articleId)
        #todo else?

    def MoveFeed(self,session,groupId):
        assert session.feedId, 'Needs current feed/group'   #todo else create initial group...

        session.lock.acquire()
        try:
            exec1=session.conn.cursor()
            exec1.execute("UPDATE FEEDS SET GROUP_ID=%d WHERE FEED_ID=%d" % (int(groupId),int(session.feedId)))
            exec1.close()

            debug('Moved feed %d to group %d' % (int(session.feedId),int(groupId)))

            session.conn.commit()
        finally:
            session.lock.release()

        session.feedlist=None

    def DeleteFeed(self,session):
        assert session.feedId, 'Needs current feed/group'   #todo else create initial group...

        session.lock.acquire()
        try:
            exec1=session.conn.cursor()
            exec1.execute("DELETE FROM FEEDS WHERE FEED_ID=%d" % int(session.feedId))
            exec1.close()
            #Note: articles are cascaded

            debug('Deleted feed %d' % int(session.feedId))

            session.conn.commit()
        finally:
            session.lock.release()

        session.feedlist=None


    def AddGroup(self,session,groupName):
        assert session.feedId, 'Needs current feed/group'   #todo else create initial group...

        session.lock.acquire()
        try:
            exec1=session.conn.cursor()
            exec1.execute("INSERT INTO GROUPS (SITE_ID, GROUP_NAME) VALUES (?, ?)" , (session.reader.siteId, strEncode(groupName)))
            exec1.close()

            debug('Added %s' % (groupName))

            session.conn.commit()
        finally:
            session.lock.release()

        session.feedlist=None


    def UpdateGroup(self,session,groupName):
        assert session.feedId, 'Needs current feed/group'   #todo else create initial group...

        session.lock.acquire()
        try:
            results1=session.conn.cursor()
            results1.execute('SELECT * FROM FEEDS WHERE FEED_ID=%d' % int(session.feedId))
            FEEDS = IMetaRow(results1.description)
            f=results1.fetchone()
            if f:
                exec1=session.conn.cursor()
                exec1.execute("UPDATE GROUPS SET GROUP_NAME=? WHERE GROUP_ID=?" , (strEncode(groupName), FEEDS(f).fields.GROUP_ID))
                exec1.close()

                debug('Updated group %d to %s' % (FEEDS(f).fields.GROUP_ID, groupName))

                session.conn.commit()
            #todo else error!
            results1.close()
        finally:
            session.lock.release()

        session.feedlist=None


    def DeleteGroup(self,session):
        assert session.feedId, 'Needs current feed/group'   #todo else create initial group...

        session.lock.acquire()
        try:
            results1=session.conn.cursor()
            results1.execute('SELECT * FROM FEEDS WHERE FEED_ID=%d' % int(session.feedId))
            FEEDS = IMetaRow(results1.description)
            f=results1.fetchone()
            if f:
                exec1=session.conn.cursor()
                exec1.execute("DELETE FROM GROUPS WHERE GROUP_ID=%d" % int(FEEDS(f).fields.GROUP_ID))
                exec1.close()
                #Note: feeds and articles are cascaded

                debug('Deleted group %d' % FEEDS(f).fields.GROUP_ID)

                session.conn.commit()
            #todo else error!
            results1.close()
        finally:
            session.lock.release()

        session.feedlist=None

    def ExportGroup(self,session):
        assert session.feedId, 'Needs current feed/group'   #todo else create initial group...

        session.lock.acquire()
        try:
            results1=session.conn.cursor()
            results1.execute('SELECT * FROM FEEDS WHERE FEED_ID=%d' % int(session.feedId))
            FEEDS = IMetaRow(results1.description)
            f=results1.fetchone()
            if f:
                results2=session.conn.cursor()
                results2.execute("SELECT * FROM GROUPS WHERE GROUP_ID=%d" % int(FEEDS(f).fields.GROUP_ID))
                GROUPS = IMetaRow(results2.description)
                g=results2.fetchone()
                if g:
                    scr=file("%s.csv" % GROUPS(g).fields.GROUP_NAME, "w")
                    writer = csv.writer(scr)
                
                    results3=session.conn.cursor()
                    results3.execute("SELECT * FROM FEEDS WHERE GROUP_ID=%d" % int(FEEDS(f).fields.GROUP_ID))
                    FEEDS2 = IMetaRow(results3.description)
                    for f3 in results3:
                        if FEEDS2(f3).fields.FEED_SUMMARY_SIZE:
                            writer.writerow( (strSafe(FEEDS2(f3).fields.FEED_NAME), strSafe(FEEDS2(f3).fields.FEED_URL), int(FEEDS2(f3).fields.FEED_SUMMARY_SIZE)) )
                        else:
                            writer.writerow( (strSafe(FEEDS2(f3).fields.FEED_NAME), strSafe(FEEDS2(f3).fields.FEED_URL), None) )

                    results3.close()
                    scr.close()

                results2.close()

                debug('Exported group %d' % FEEDS(f).fields.GROUP_ID)
            #todo else error!
            results1.close()
        finally:
            session.lock.release()

    def ImportGroup(self,session,groupName):
        assert session.feedId, 'Needs current feed/group'   #todo else create initial group...

        session.lock.acquire()
        try:
            #somehow get the filename
            try:
                scr=file("%s.csv" % groupName)
            except:
                self.showAdminForm(session, session.feedId)
                raise
            
            #If groupName already exists, reject the import  todo:find its id and use that to merge?
            results1=session.conn.cursor()
            try:
                results1.execute("SELECT * FROM GROUPS WHERE GROUP_NAME='%s'" % strSafe(groupName))
                if results1.fetchone():
                    self.showAdminForm(session, session.feedId)
                    raise 'Group already exists'
            finally:
                results1.close()

            reader = csv.reader(scr)

            exec1=session.conn.cursor()
            exec1.execute("INSERT INTO GROUPS (SITE_ID, GROUP_NAME) VALUES (?, ?)" , (session.reader.siteId, strEncode(groupName)))
            exec1.close()

            debug('Imported %s' % (groupName))
            
            for row in reader:
                exec1=session.conn.cursor()
                if row[2]:
                    exec1.execute("INSERT INTO FEEDS (GROUP_ID, FEED_NAME, FEED_URL, FEED_SUMMARY_SIZE) VALUES (LATEST_SEQUENCE(GROUPS), '%s', '%s', %d);\n" % (strSafe(row[0]), strSafe(row[1]), int(row[2])) )
                else:                    
                    exec1.execute("INSERT INTO FEEDS (GROUP_ID, FEED_NAME, FEED_URL) VALUES (LATEST_SEQUENCE(GROUPS), '%s', '%s');\n" % (strSafe(row[0]), strSafe(row[1])) )
                exec1.close()           
        
            session.conn.commit()                
        finally:
            session.lock.release()

        session.feedlist=None


    def showStatisticsForm(self,session):
        self.showFormHead(session,'actionStatisticsForm')

        gm=self.getGroupMenu(session, session.feedId)
        feedId=gm[0]
        feedName=gm[1]
        feedSummary_size=gm[2]
        feeds=gm[3]

        feeds.append('<br />')

        self.wfile.write( FORM_ADDFEED_TOP % {'title':session.reader.siteName, 'date':datetime.today().strftime(USER_DATE), 'feeds':''.join(feeds)} )

        stats=[]
        session.lock.acquire()
        try:
            results1=session.conn.cursor()
            results1.execute("""SELECT FEED_ID, (SELECT FEED_NAME FROM FEEDS WHERE FEEDS.FEED_ID=ARTICLES.FEED_ID) AS NAME, 
                              COUNT(*) AS NUMBER, MIN(ARTICLE_DATE) AS EARLIEST, SUM(OCTET_LENGTH(ARTICLE_DESCRIPTION)) AS TOTAL_SIZE FROM ARTICLES
                              GROUP BY FEED_ID ORDER BY NUMBER DESC""")
            STATS = IMetaRow(results1.description)
            for s in results1:
                stats.append(FORM_STATS % (STATS(s).fields.NAME, STATS(s).fields.NUMBER, STATS(s).fields.EARLIEST.strftime(USER_DATE), STATS(s).fields.TOTAL_SIZE))

            results1.close()
        finally:
            session.lock.release()

        self.wfile.write( FORM_STATS_MAIN % {'title':'Article statistics', 'stats':''.join(stats)} )

        self.showFormFoot(session)


    def actionStatisticsForm(self,session,fields):
        command=self.getCommand(session,fields)

        if command == "Main":
            self.showMainForm(session, session.feedId, session.articleId)
        #todo else?

    def UpdateServer(self,session,serverPurgeDays):
        if serverPurgeDays:
            try:
                serverPurgeDays=int(serverPurgeDays)
                session.lock.acquire()
                try:
                    exec1=session.conn.cursor()
                    exec1.execute("UPDATE SITES SET ARTICLE_PURGE_DAYS=? WHERE SITE_ID=?" , (serverPurgeDays, session.reader.siteId))
                    exec1.close()
        
                    session.conn.commit()
        
                    session.reader.article_purge_days=serverPurgeDays
        
                    debug('Updated site %d purge days to %d' % (session.reader.siteId, serverPurgeDays))          
                finally:
                    session.lock.release()
            except:
                serverPurgeDays=None               

    def Purge(self,session):
        session.lock.acquire()
        try:
            results1=session.conn.cursor()
            base=datetime.today()
            seq=timedelta(days=session.reader.article_purge_days)
            base=base-seq
            cutoff_date=base.strftime(ISO_DATE)
            results1.execute("SELECT COUNT(*) AS TOTAL FROM ARTICLES WHERE ARTICLE_DATE<TIMESTAMP'%s'" % cutoff_date)
            ARTICLES = IMetaRow(results1.description)
            a=results1.fetchone()
            if a:
                print '  Purging %d articles before %s...' % (ARTICLES(a).fields.TOTAL, cutoff_date)

                exec1=session.conn.cursor()
                exec1.execute("DELETE FROM ARTICLES WHERE ARTICLE_DATE<TIMESTAMP'%s'" % cutoff_date)
                exec1.close()

                session.conn.commit()
            #todo else error!
            results1.close()
        finally:
            session.lock.release()

    def showAdminForm(self, session, feedId=None):
        self.showFormHead(session,'actionAdminForm')

        gm=self.getGroupMenu(session, feedId)
        feedId=gm[0]
        feedName=gm[1]
        feedSummary_size=gm[2]
        feeds=gm[3]

        feeds.append('<br />')

        self.wfile.write( FORM_ADDFEED_TOP % {'title':session.reader.siteName, 'date':datetime.today().strftime(USER_DATE), 'feeds':''.join(feeds)} )

        session.lock.acquire()
        try:
            if feedId:
                results1=session.conn.cursor()
                results1.execute('SELECT * FROM FEEDS NATURAL JOIN GROUPS WHERE FEED_ID=%d' % int(feedId))
                FEEDGROUPS = IMetaRow(results1.description)
                f=results1.fetchone()
                if f:
                    feedSummarysize=FEEDGROUPS(f).fields.FEED_SUMMARY_SIZE
                    if not feedSummarysize:
                        feedSummarysize=''
                    self.wfile.write( FORM_ADMIN_MAIN % ('Admin', FEEDGROUPS(f).fields.FEED_NAME, FEEDGROUPS(f).fields.FEED_URL, feedSummarysize, FEEDGROUPS(f).fields.GROUP_NAME, str(session.reader.article_purge_days)) )
                else:
                    self.wfile.write( FORM_ADMIN_MAIN % ('Admin', '', '', '', '', str(session.reader.article_purge_days)) )
                results1.close()
            else:
                self.wfile.write( FORM_ADMIN_MAIN % ('Admin', '', '', '', '', str(session.reader.article_purge_days)) )
        finally:
            session.lock.release()

        self.showFormFoot(session)

    def actionAdminForm(self,session,fields):
        command=self.getCommand(session,fields)

        if command == "Main":
            self.showMainForm(session, session.feedId, session.articleId)
        elif command == "Update Group":
            if fields.has_key('groupname'):
                gName=fields['groupname']
            self.UpdateGroup(session, gName)
            self.showMainForm(session, session.feedId, session.articleId)
        elif command == "Delete Group":
            self.DeleteGroup(session)
            self.showMainForm(session, None, None)
        elif command == "Export Group":
            self.ExportGroup(session)
            self.showMainForm(session, session.feedId, session.articleId)
        elif command == "Update Feed":
            if fields.has_key('feedname'):
                fName=fields['feedname']
            if fields.has_key('feedurl'):
                fURL=fields['feedurl']
            if fields.has_key('feedsummarysize'):
                fsummarysize=fields['feedsummarysize']
            self.UpdateFeed(session, fName, fURL, fsummarysize)
            self.showMainForm(session, session.feedId, session.articleId)
        elif command == "Move Feed...":
            self.showMoveFeedForm(session)
        elif command == "Delete Feed":
            self.DeleteFeed(session)
            self.showMainForm(session, None, None)
        elif command == "Add Group":
            if fields.has_key('newgroupname'):
                gName=fields['newgroupname']
            self.AddGroup(session, gName)
            self.showMainForm(session, session.feedId, session.articleId)
        elif command == "Import Group":
            if fields.has_key('newgroupname'):
                gName=fields['newgroupname']
            self.ImportGroup(session, gName)
            self.showMainForm(session, session.feedId, session.articleId)
        elif command == "Statistics":
            self.showStatisticsForm(session)
        elif command == "Update Server":
            if fields.has_key('serverpurgedays'):
                fpurgedays=fields['serverpurgedays']
            self.UpdateServer(session, fpurgedays)        
            self.showMainForm(session, session.feedId, session.articleId)
        elif command == "Purge":
            self.Purge(session)
            self.showMainForm(session, session.feedId, session.articleId)
        elif command == "Shutdown":
            if session.state==STATE_CONNECTED:
                session.conn.close()

            debug('Stopping ThinkSQL server...')
            print "Stopping ThinkSQL server..."

            import telnetlib
            tn = telnetlib.Telnet(session.host,session.port)
            tn.write("\r\n");
            tn.read_until("connections",2);
            tn.write("shutdown;\r\n")
            print tn.read_eager()
            tn.close()

            self.wfile.write(PAGE_HEAD)
            self.wfile.write(FORM_HEADER_TOP % {'title':session.reader.siteName, 'date':datetime.today().strftime(USER_DATE)} )
            self.wfile.write('>Server shutdown')
            self.wfile.write(PAGE_FOOT)

            global httpd
            httpd.proceed = 0

        #todo else?

    def do_GET(self):
        if self.path == "/default.css":
            self.wfile.write( file('default.css').read() )
        elif self.path == "/closed.jpg":
            self.wfile.write( open('closed.jpg', 'rb').read() )
        elif self.path == "/open.jpg":
            self.wfile.write( open('open.jpg', 'rb').read() )
        elif self.path == "/bullet.gif":
            self.wfile.write( open('bullet.gif', 'rb').read() )
        elif self.path == "/rss.gif":
            self.wfile.write( open('rss.gif', 'rb').read() )
        elif self.path == "/raven.gif":
            self.wfile.write( open('raven.gif', 'rb').read() )
        elif self.path[:2] == "/?":
            self.respond()
        else:
            self.respond()

    def respond(self):
        """Called for each request"""
        try:
            self.send_response(200, 'OK')
            self.send_header("Content-type", "text/html")
            self.end_headers()
    
            #Parse the form fields
            fields = cgi.parse_qs(self.path[2:], True)
            for k in fields.keys():
                fields[k] = fields[k][0] # Convert from lists to individual items
    
            #Find current session or create a new one
            global sessions
            session=None
            if fields.has_key('session_id'):
                current_session_id=int(fields['session_id'])
                if current_session_id in sessions:
                    session=sessions[current_session_id]
                else:
                    debug('Session has expired... reconnecting...')
            if session==None:
                session=ServerSession()
                sessions[session.id]=session
    
            #Index (or missing command => restart)
            if session.state==STATE_DISCONNECTED:
                debug('Connecting to %s %d as %s' % (session.host, session.port, session.username))
                try:
                    session.lock.acquire()
                    try:
                        global singleUser
                        if singleUser:
                            global singleConn
                            if not singleConn:
                                singleConn=session.conn=ThinkSQL.connect(host=session.host, port=session.port, user=session.username, password=session.password) #todo database=catalog
                            session.conn=singleConn
                        else:
                            session.conn=ThinkSQL.connect(host=session.host, port=session.port, user=session.username, password=session.password) #todo database=catalog
                        session.state=STATE_CONNECTED
                        session.reader=feedReader(session.conn)
                    finally:
                        session.lock.release()
                except Exception, e:
                    self.showError(session,'Failed to connect: %s' % str(e))
                    raise
    
                self.showMainForm(session)
            else:
                if fields.has_key('page'):
                    if hasattr(self,fields['page']):
                        session.page=fields['page']
                        pageRef=getattr(self,fields['page'])
                        pageRef(session,fields)
                    else:
                        self.showError(session,'Unknown page: ' % fields['page'])
                #todo else?
        except Exception, err:
            if err[0]<>errno.ECONNABORTED:
                raise
            #else thread overlapping output to wfile = ok
            
    def log_request(self, code):
       pass


BaseSClass = BaseHTTPServer.HTTPServer
#Multithreaded server: class testServer(SocketServer.ThreadingMixIn, BaseSClass):
class testServer(BaseSClass):
    def serve_forever(self):
        """Handle one request at a time until doomsday... almost."""
        self.proceed = 1
        while self.proceed:
            self.handle_request()

debugSwitch=False
singleUser=True     #turn off to allow multiple connections

# When run as a script we start the server on port 8000:
if __name__ == '__main__':
    print "Starting Raven server..."
    if sys.argv[1:]:
        port = int(sys.argv[1])
    else:
        port = 8000
    
    server_address = ('', port)

    httpd = testServer(server_address, testRequestHandler)

    sa = httpd.socket.getsockname()    

    debug("Finding SQL server...")
    try:
        alreadyRunning=ThinkSQL.connect(host=DEFAULT_HOST, port=DEFAULT_PORT, user=DEFAULT_USER, password='') #todo database=catalog
        alreadyRunning.close()
    except:
        debug("Starting ThinkSQL server...")
        print "Starting ThinkSQL server..."
        import os
        if os.name=='posix':
            os.system('./ThinkSQL raven&')
        else:
            os.system('start "" "thinksql" raven')     #add prefix of c:\\program files\\thinksql\\ if ThinkSQL server already available

    if singleUser:
        singleConn=None

    debug("Serving HTTP on %s port %s" % (sa[0], str(sa[1])))
    print "Serving HTTP on", sa[0], "port", sa[1]
    try:
        httpd.serve_forever()
    finally:
        debug('Closed down')
        print "Closed down"
